"""The most common steps in the release of a Python package."""

import requests

from bag.check_rst import check_rst_file
from bag.console import bool_input
from bag.pathlib_complement import Path

from . import ReleaseStep, StopRelease, CommandStep
from .regex import version_in_python_source_file

__all__ = (
    "Shell",
    "CheckRstFiles",
    "InteractivelyApprovePackage",
    "CheckTravis",
    "InteractivelyEnsureChangesDocumented",
    "SetVersionNumberInteractively",
    "TwineUploadSource",
    "TwineUploadWheel",
    "SetFutureVersion",
    "ErrorStep",
    "Warn",
)


class Shell(ReleaseStep):
    """Run some shell command."""

    ERROR_CODE = 2

    def __init__(self, command, stop_on_failure=True):  # noqa
        self.COMMAND = command
        self.stop_on_failure = stop_on_failure
        self.no_rollback = "Unable to roll back the step {0}".format(self)

    def __call__(self):  # noqa
        self._execute_or_complain(self.COMMAND)  # sets self.success

    def __str__(self):
        return "[" + self.COMMAND + "]"


class CheckRstFiles(ReleaseStep):
    """Helps keep documentation correct by verifying .rst files.

    If files are not provided to the constructor, recursively finds
    *.rst files in the current directory and subdirectories.
    """

    ERROR_CODE = 4

    def __init__(self, *files):  # noqa
        self.paths = files

    def __call__(self):  # noqa
        paths = self.paths or Path(".").walk(filter=lambda p: p.suffix == ".rst")
        for path in paths:
            self.log.info(f"Checking {path}")
            warnings = check_rst_file(str(path))
            if warnings:
                raise StopRelease(
                    "There are errors in {0}:\n{1}".format(
                        path, "\n".join([str(w) for w in warnings])
                    )
                )


class InteractivelyApprovePackage(ReleaseStep):
    """Ask the user to manually verify source and wheel files."""

    ERROR_CODE = 5

    def __call__(self):  # noqa
        # TODO: Optionally xdg-open the archive for convenience
        print("The new packages have been generated in the dist/ directory.")
        print("Since they are really just zip files, you should now open them")
        print("and check that no desired source files are missing.")
        if not bool_input("Do you approve the archive contents?"):
            raise StopRelease("Package content not approved.")


class InteractivelyEnsureChangesDocumented(ReleaseStep):
    """Step that just bugs the user to verify the CHANGES file."""

    ERROR_CODE = 3

    def __call__(self):  # noqa
        if bool_input("Did you remember to update the CHANGES file?"):
            self.log.debug("User says CHANGES file is up to date.")
        else:
            raise StopRelease("One more joeshlabotniked release is avoided.")


class CheckTravis(ReleaseStep):
    """Check the status, on travis-ci.org, of the latest build."""

    ERROR_CODE = 91
    URL = "https://api.travis-ci.org/repos/{github_user}/{github_repository}/builds"

    def __call__(self):  # noqa
        branch = self.config.get("branch", "master")
        resp = requests.get(self.URL.format(**self.config))
        builds = resp.json()
        onbranch = filter(lambda x: x["branch"] == branch, builds)
        finished = list(filter(lambda x: x["state"] == "finished", onbranch))
        if len(finished) == 0:
            raise StopRelease('Travis has not built branch "{0}" yet.'.format(branch))
        latest = finished[0]
        if latest["result"] == 0:
            self.log.info(
                'No problem in latest Travis build: "{0}"'.format(latest.get("message"))
            )
            self._succeed()
        else:
            raise StopRelease(
                'Last Travis build on branch "{0}" failed.'.format(branch)
            )


class SetVersionNumberInteractively(ReleaseStep):
    """Ask user for the new version number and write it on the source code."""

    ERROR_CODE = 6

    def __call__(self):  # noqa
        releaser = self.releaser
        path = releaser.config["version_file"]
        keyword = releaser.config.get("version_keyword", "version")
        releaser.old_version = version_in_python_source_file(path, keyword=keyword)
        print("Current version: {0}".format(releaser.old_version))
        releaser.the_version = input("What is the new version number? ")
        # Write the new version onto the source code
        version_in_python_source_file(
            path, keyword=keyword, replace=releaser.the_version
        )
        self._succeed()


class TwineUploadSource(CommandStep):
    """Use *twine* to upload a source distribution to pypi."""

    ERROR_CODE = 8
    no_rollback = "Cannot roll back the sdist upload to http://pypi.python.org"

    def __call__(self):  # noqa
        name = self.config["github_repository"]
        version = self.releaser.the_version
        self._execute_expect_zero(f"twine upload dist/{name}-{version}.tar.gz")

    def _validate_command_output(self, command_output):
        return "Server response (200): OK" in command_output


class TwineUploadWheel(CommandStep):
    """Use *twine* to upload a wheel to pypi."""

    ERROR_CODE = 11
    no_rollback = "Cannot roll back the wheel upload to http://pypi.python.org"

    def __call__(self):  # noqa
        name = self.config["github_repository"]
        version = self.releaser.the_version
        self._execute_expect_zero(
            f"twine upload dist/{name}-{version}-py3-none-any.whl"
        )

    def _validate_command_output(self, command_output):
        return "Server response (200): OK" in command_output


class SetFutureVersion(ReleaseStep):
    """Set the development version number in source code after release."""

    ERROR_CODE = 9

    def __call__(self):  # noqa
        releaser = self.releaser
        path = releaser.config["version_file"]
        keyword = releaser.config.get("version_keyword", "version")
        # If the SetVersionNumberInteractively step is disabled for debugging,
        # we can still execute the current step, by populating the_version:
        if releaser.the_version is None:
            releaser._the_version = version_in_python_source_file(path, keyword=keyword)
        self.log.info(
            "Ready for the next development cycle! Setting version "
            + releaser.future_version
        )
        version_in_python_source_file(
            path, keyword=keyword, replace=releaser.future_version
        )
        self._succeed()


class ErrorStep(CommandStep):
    """Raise an exception to force a rollback. Good for testing."""

    COMMAND = "thisCommandDontExist"
    ERROR_CODE = 255


class Warn(ReleaseStep):
    """Just print a warning on the screen and on the log file."""

    def __init__(self, msg: str):  # noqa
        self.msg = msg

    def __call__(self):  # noqa
        self.log.warn(self.msg)
        self.success = True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['releaser']

package_data = \
{'': ['*']}

install_requires = \
['bag>=5.0.0,<6.0.0',
 'docutils<0.18',
 'grimace>=0.1.0,<0.2.0',
 'requests',
 'wheel']

setup_kwargs = {
    'name': 'releaser',
    'version': '3.0.0',
    'description': 'Automates the process of releasing a new version of some software.',
    'long_description': None,
    'author': 'Nando Florestan',
    'author_email': 'nandoflorestan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nandoflorestan/releaser',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Framework for releasing Python software without forgetting steps."""

import subprocess
from sys import platform
from pkg_resources import parse_version
from bag.console import bool_input, screen_header
from bag.log import setup_log
from .regex import error_in_version


class StopRelease(RuntimeError):
    """Release steps should raise this exception to stop the whole program."""


class ReleaseStep:
    """Abstract base class for release steps."""

    ERROR_CODE = 1
    success = None
    stop_on_failure = True

    def __call__(self):
        """Override this method to do the main work of the release step.

        On error, StopRelease should be raised.
        """
        raise StopRelease("Step not implemented.")

    def _succeed(self):
        self.success = True

    def _fail(self, msg):
        self.success = False
        if self.stop_on_failure:
            raise StopRelease(msg)
        else:
            self.log.warning(msg + "\nContinuing anyway.")

    def _execute(self, command, input="", shell=True):
        p = subprocess.Popen(
            command,
            shell=shell,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=not platform.startswith("win"),
        )
        i, o, e = (p.stdin, p.stdout, p.stderr)
        if input:
            i.write(input)
        i.close()
        return_code = p.wait()
        normal_output = o.read().strip().decode("utf-8")
        error_output = e.read().strip().decode("utf-8")
        o.close()
        e.close()
        if normal_output:
            self.log.debug(normal_output)
        if error_output:
            self.log.error(error_output)
        return return_code, normal_output + error_output

    def _execute_or_complain(
        self,
        command,
        input="",
        shell=True,
        msg="Command failed with code {code}: {cmd}",
    ):
        return_code, text = self._execute(command, input, shell)
        if return_code == 0:
            if hasattr(self, "_validate_command_output"):
                if self._validate_command_output(text):
                    self._succeed()
                else:
                    self._fail(msg.format(code=return_code, cmd=command))
            else:
                self._succeed()
        else:
            self._fail(msg.format(code=return_code, cmd=command))
        return text

    def _execute_expect_zero(
        self,
        command,
        input="",
        shell=True,
        msg="Command failed with code {code}: {cmd}",
    ):
        return_code, text = self._execute(command, input, shell)
        if return_code == 0:
            self._succeed()
        else:
            self._fail(msg.format(code=return_code, cmd=command))
        return text

    def __str__(self):
        return type(self).__name__  # good default identifier for most steps


class CommandStep(ReleaseStep):
    """Abstract base class for any release step that runs a command."""

    def __call__(self):
        self._execute_or_complain(self.COMMAND)  # sets self.success


class Releaser:
    """Class that manages the whole release process."""

    def __init__(self, config, *steps):
        self.created_tags = []
        self.config = config
        # Configure logging
        path = config.get("log_file", "release.log.utf-8.tmp")
        screen_level = config.get("verbosity", "info")
        self.log = setup_log(
            path=path,
            rotating=False,
            file_mode="w",
            disk_level="debug",
            screen_level=screen_level,
        )
        # Convert steps provided by the user into real instances
        self.instances = []
        for step in steps:
            if isinstance(step, type):
                step = step()  # Instantiate the ReleaseStep subclass
            step.config = config
            step.releaser = self
            step.log = self.log
            self.instances.append(step)
        # Now that all steps are correctly instantiated,
        # it is safe to start running them by calling release().

    def release(self):
        self.rewindable = []
        self.non_rewindable = []
        for step in self.instances:
            self.log.info(screen_header(step))
            try:
                step()
            except StopRelease as e:
                self.log.critical(
                    "Release process stopped at step {0}:\n{1}".format(step, e)
                )
                self.rewind()
                from sys import exit

                exit(step.ERROR_CODE)
            except Exception:
                self.log.debug("Noooo!!!", exc_info=True)
                self.rewind()
                raise
            else:
                if step.success and hasattr(step, "no_rollback"):
                    self.non_rewindable.append(step.no_rollback)
                elif step.success and hasattr(step, "rollback"):
                    self.rewindable.append(step)
        self.log.info(
            "Successfully released version {0}. "
            "Sorry for the convenience, mcdonc!".format(self.the_version)
        )

    def rewind(self):
        say = self.log.critical
        say("\n" + screen_header("ROLLBACK", decor="*"))
        # Some steps offer a warning because they cannot be rolled back.
        say("\n".join(self.non_rewindable))  # Display them.
        if not self.rewindable:
            say("No steps to roll back, but the release process FAILED.")
            return
        steps = list(reversed(self.rewindable))
        print(
            "I am about to roll back the following steps:\n{0}".format(
                ", ".join([str(step) for step in steps])
            )
        )
        if not bool_input("Continue?", default=True):
            return
        for step in steps:
            self.log.critical(screen_header("ROLLBACK {0}".format(step)))
            try:
                step.rollback()
            except Exception as e:
                self.log.error(
                    "Could not roll back step {0}:\n{1}".format(step, str(e))
                )

    _old_version = None  # 0.1.2dev (exists when the program starts)
    _the_version = None  # 0.1.2    (the version being released)

    @property
    def old_version(self):
        return self._old_version

    @old_version.setter
    def old_version(self, val):
        self._old_version = val
        self.log.debug("Old version: {0}".format(val))

    @property
    def the_version(self):
        return self._the_version

    @the_version.setter
    def the_version(self, val):
        val = val.strip()
        e = error_in_version(val, allow_dev=False)
        if e:
            raise StopRelease(e)
        if parse_version(self.old_version) >= parse_version(val):
            raise StopRelease(
                "No, the version number must be higher than the current one!"
            )
        self._the_version = val
        self.log.debug("Version being released: {0}".format(val))

    @property
    def future_version(self):  # 0.1.3.dev1 (development version after release)
        parts = self.the_version.split(".")
        parts[-1] = str(int(parts[-1]) + 1)
        return ".".join(parts) + ".dev1"
